
#include "vmu_pkg.h"
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <zlib.h>

#define READ_BYTE(BASE, ADDR) (BASE)[(ADDR)^1]
#define READ_WORD(BASE, ADDR) (((BASE)[(ADDR)+1]<<8) |          \
                              (BASE)[(ADDR)])
#define READ_LONG(BASE, ADDR) (((BASE)[(ADDR)+1]<<24) |         \
                               ((BASE)[(ADDR)+0]<<16) |      \
                               ((BASE)[(ADDR)+3]<<8) |       \
                                (BASE)[(ADDR)+2])

static uint8_t block[0x4000];
uint8_t cart_rom[0x400000];
int rom_size;
uint32_t rom_crc;

void deinterleave_block(uint8_t *src)
{
    int i;
    memcpy(block, src, 0x4000);
    for(i = 0; i < 0x2000; i += 1)
    {
        src[i*2+0] = block[0x2000 + (i)];
        src[i*2+1] = block[0x0000 + (i)];
    }
}

void bswap(uint8_t *mem, int length)
{
    int i;
    for(i = 0; i < length; i += 2)
    {
        uint8_t temp = mem[i+0];
        mem[i+0] = mem[i+1];
        mem[i+1] = temp;
    }
}

int load_rom(char *filename)
{
    int size, offset = 0;
    uint8_t header[0x200];
    uint8_t *ptr;

    FILE *f = fopen(filename, "rb");
    if(!f) {
    	printf("Can not open ROM: %s\n", filename);
    	return 0;
    }
    fseek(f, 0, SEEK_END);
    size = ftell(f);
    fseek(f, 0, SEEK_SET);
    ptr = malloc(size);
    fread(ptr, size, 1, f);
    fclose(f);

    rom_size = size;

    if((size / 512) & 1)
    {
        int i;

        size -= 512;
        offset += 512;

        memcpy(header, ptr, 512);

        for(i = 0; i < (size / 0x4000); i += 1)
        {
            deinterleave_block(ptr + offset + (i * 0x4000));
        }
    }

    memset(cart_rom, 0, 0x400000);
    if(size > 0x400000) size = 0x400000;
    memcpy(cart_rom, ptr + offset, size);

    /* Free allocated file data */
    free(ptr);

    /* Calculate CRC32 */
    rom_crc = crc32(0L, Z_NULL, 0);
    rom_crc = crc32(rom_crc, cart_rom, rom_size);

    /* Byteswap the ROM */
    bswap(cart_rom, 0x400000);

    return (1);
}

uint8_t cart_ram[128 * 1024];
uint8_t sramfile_buffer[128 * 1024];
int sramfile_offset;
uint32_t sram_start, sram_end, sram_present;

void sram_init()
{
	/* Check the cart header */
	if(READ_WORD(cart_rom, 0x0001B0) == 0x5241)
	{
		sram_present = 1;
		sram_start = READ_LONG(cart_rom, 0x0001B4) & 0xFFFFFF;
		sram_end = READ_LONG(cart_rom, 0x0001B8) & 0xFFFFFF;
	}
	else
	{
		/* No SRAM in the header, so assume it's 64K (for S3K) */
		sram_present = 0;
		sram_start = 0x200000;
		sram_end = 0x20FFFF;
	}
}

void sram_write_uint8(uint8_t value)
{
	sramfile_buffer[sramfile_offset++] = value;
}

void sram_write_uint16(uint16_t value)
{
	memcpy(sramfile_buffer + sramfile_offset, &value, sizeof(value));
	sramfile_offset += sizeof(value);
}

void sram_write_uint32(uint32_t value)
{
	memcpy(sramfile_buffer + sramfile_offset, &value, sizeof(value));
	sramfile_offset += sizeof(value);
}

/* Try to save SRAM */
void sram_save()
{
	int addr_odd, addr_even;
	uint16_t addr_first, addr_last;
	uint8_t addr_mask;
	int i;

	// Work out what data needs to be stored
	addr_first = 0xFFFF;
	addr_last = 0;
	addr_odd = addr_even = 0;
	for(i = 0; i < 0x10000; i++)
	{
		if(cart_ram[i])
		{
			if(addr_first > i)
				addr_first = i;
			if(addr_last < i)
				addr_last = i;
			if(i & 0x01)
				addr_odd = 1;
			else
				addr_even = 1;
		}
	}
	addr_mask = 0;
	if(addr_even)
		addr_mask |= 0x01;
	if(addr_odd)
		addr_mask |= 0x02;

	// Write header
	sram_write_uint32(sram_start);
	sram_write_uint32(sram_end);
	sram_write_uint16(addr_first);
	sram_write_uint16(addr_last);
	sram_write_uint8(addr_mask);

	// Write data
	for(i = addr_first; i <= addr_last; i++)
	{
		if(i & 0x01)
		{
			if(addr_mask & 0x02)
				sram_write_uint8(cart_ram[i]);
		}
		else
		{
			if(addr_mask & 0x01)
				sram_write_uint8(cart_ram[i]);
		}
	}
}

void sram_write_close()
{
	static uint8_t compress_buffer[64*1024];
	uLongf destLen = sizeof(compress_buffer);

	char filename[14];
	sprintf(filename, "%08X.gps", (unsigned int)rom_crc);

	/* Compress data */
	if(compress2(compress_buffer, &destLen, sramfile_buffer, sramfile_offset, 9) != Z_OK)
	{
		printf("Can not compress\n");
		return;
	}

	/* Prepare a VMU package */
	vmu_pkg_t vmupkg;
	memset(&vmupkg, 0, sizeof(vmupkg));
	strncpy(vmupkg.desc_short, "Genesis Plus SRAM", 20);
	strncpy(vmupkg.desc_long, (char *)(cart_rom + 0x150), 36);
	bswap((uint8_t *)vmupkg.desc_long, sizeof(vmupkg.desc_long));
	strncpy(vmupkg.app_id, "Genesis Plus / DC", 20);
	vmupkg.icon_cnt = 0;
	vmupkg.data_len = destLen;
	vmupkg.data = compress_buffer;

	/* Package VMU file */
	uint8_t *vmufile_raw;
	size_t vmufile_size;
	vmu_pkg_build(&vmupkg, &vmufile_raw, &vmufile_size);

	/* Write the data to the VMU */
	FILE * f = fopen(filename, "wb");
	if(!f) {
		printf("Can not open %s\n", filename);
		free(vmufile_raw);
		return;
	}
	fwrite(vmufile_raw, vmufile_size, 1, f);
	fclose(f);

	/* Free allocated data */
	free(vmufile_raw);

	printf("VMU File Written\n");
}


int main(int argc, char *argv[]) {
	memset(cart_ram, 0, sizeof(cart_ram));
	memset(sramfile_buffer, 0, sizeof(sramfile_buffer));
	sramfile_offset = 0;

	if(argc < 2) {
		printf("Usage: convert_save rom.bin input.srm\n");
		return 1;
	}

	// Load the ROM
	if(!load_rom(argv[1])) {
		printf("Could not load ROM %s\n", argv[1]);
		return 1;
	}

	// Init SRAM from header
	sram_init();

	printf("sram_start: 0x%08X\n", sram_start);
	printf("sram_end: 0x%08X\n", sram_end);
	printf("CRC: %08x\n", rom_crc);

	// Read SRAM file
	FILE *f = fopen(argv[2], "rb");
	if(!f) {
		printf("Can't open %s\n", argv[2]);
		return 1;
	}
	fseek(f, 0, SEEK_END);
	size_t len = ftell(f);
	fseek(f, 0, SEEK_SET);
	fread(cart_ram, len, 1, f);
	fclose(f);

	// Convert to GP/DC's internal format
	sram_save();

	// Save in a compressed VMU package file
	sram_write_close();



	return 0;
}
